/*
	SmoothD plugin for AviSynth -- 
		This filter deblocks the input frames while trying to preserve detail.
		Only works with YV12 input.

	Copyright (C) 2004 Tobias Bergmann
  
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.
	 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
		
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The author can be contacted at:
	tobias.bergmann at bergdichter dot de

history:

V0.0.1: 1st compile

V0.0.2: parameter passing working; YV12 input working

V0.0.3:	memory management working; zero_weight working (12h)

V0.0.4: shift_offsets determined; num_shift range corrected (13h)

V0.0.5: proprietary code removed; xvid source added; documentation html written (16h)

V0.0.6: color bug fixed (17h)

V0.0.7: flat quant working; shift working; border bug fixed; PROCESS_UV added (23.5h)

V0.0.8: adaptive_shift working; add 6of9 HVS matrix; precompute Qp[]; blacken borders; flat8 DCT (30.5h)

V0.0.9: use xvid quant code; working Flat8 matrix; working 6of9HVS matrix (35h)

future plan:

V0.1:	working U and V

  use new instead of malloc
  make codepaths for C, MMX, SSE and SSE2
  exchange x (cx) loops by memcopies or even asm subroutines
  optimize blacken borders
  speed optimizations (maybe combine dct+quant+idct)
  try once:	for (int iii=0; iii<64; iii++) Qp[iii]=min(255,quant*qm[63-iii]);
  ...

*/

#include "SmoothD.h"

SmoothD::SmoothD(PClip _child, const int _quant, const int _num_shift, const int _adaptive_shift, const int _zero_weight, IScriptEnvironment* env) : 
	GenericVideoFilter(_child), quant(_quant), num_shift(_num_shift), adaptive_shift(_adaptive_shift), zero_weight(_zero_weight) {
	// parameter checking goes in here
	if (!vi.IsYUV())
		env->ThrowError("SmoothD: input to filter must be in YUV");

	try {
	  AVSValue lim_args[1] = { child };
	  child = env->Invoke("Limiter", AVSValue(lim_args,1)).AsClip();
	} catch (AvisynthError) {}  // Just to be safe.

	vi_sizey=vi.height;
	vi_sizex=vi.width;

	// allocate akku (exact dimensions of video)
	int sizeY= vi.height*vi.width;
	int sizeU=(vi.height*vi.width)/4;

	akkuY= (uint16_t*) malloc(sizeof(uint16_t)*sizeY);
	if (!akkuY) env->ThrowError("SmoothD: out of memory");

	akkuU= (uint16_t*) malloc(sizeof(uint16_t)*sizeU);
	if (!akkuU) env->ThrowError("SmoothD: out of memory");

	akkuV= (uint16_t*) malloc(sizeof(uint16_t)*sizeU);
	if (!akkuV) env->ThrowError("SmoothD: out of memory");


	akkuY_src= (uint16_t*) malloc(sizeof(uint16_t)*sizeY);
	if (!akkuY_src) env->ThrowError("SmoothD: out of memory");

	akkuU_src= (uint16_t*) malloc(sizeof(uint16_t)*sizeU);
	if (!akkuU_src) env->ThrowError("SmoothD: out of memory");

	akkuV_src= (uint16_t*) malloc(sizeof(uint16_t)*sizeU);
	if (!akkuV_src) env->ThrowError("SmoothD: out of memory");

	// allocate akku_tmp (video dimensions + "headroom")

	int jblocks=2*((vi.height+15)/16); // minimum # of blocks to be worked on
	int iblocks=2*((vi.width +15)/16);

	sizeY=(jblocks+1)*(iblocks+1)*64;
	sizeU=(jblocks+1)*(iblocks+1)*64/4;

	akkuY_tmp= (uint16_t*) malloc(sizeof(uint16_t)*sizeY);
	if (!akkuY_tmp) env->ThrowError("SmoothD: out of memory");

	akkuU_tmp= (uint16_t*) malloc(sizeof(uint16_t)*sizeU);
	if (!akkuU_tmp) env->ThrowError("SmoothD: out of memory");

	akkuV_tmp= (uint16_t*) malloc(sizeof(uint16_t)*sizeU);
	if (!akkuV_tmp) env->ThrowError("SmoothD: out of memory");

	// define quants here
#ifdef USE_Flat8
	const uint8_t my_intra_matrix[64] = {
		 8,  8,  8,  8,  8,  8,  8,  8,
		 8,  8,  8,  8,  8,  8,  8,  8,
		 8,  8,  8,  8,  8,  8,  8,  8,
		 8,  8,  8,  8,  8,  8,  8,  8,
		 8,  8,  8,  8,  8,  8,  8,  8,
		 8,  8,  8,  8,  8,  8,  8,  8,
		 8,  8,  8,  8,  8,  8,  8,  8,
		 8,  8,  8,  8,  8,  8,  8,  8
	};
#endif
#ifdef USE_6o9HVS
	const uint8_t my_intra_matrix[64] = {
		 8, 11, 12, 12, 13, 15, 16, 17,
		11, 11, 12, 12, 14, 15, 16, 17,
		12, 12, 13, 14, 15, 16, 17, 18,
		12, 12, 14, 16, 17, 18, 19, 19,
		13, 14, 15, 17, 19, 20, 20, 20,
		15, 15, 16, 18, 20, 21, 22, 22,
		16, 16, 17, 19, 21, 22, 23, 24,
		17, 17, 18, 19, 20, 22, 24, 24
	};
#endif

	set_intra_matrix(quant_matrix, my_intra_matrix);

	// if set_intra_matrix works then qm and Qp are no longer needed
	// define quants here
	const unsigned char qm[64] = {
		 8, 11, 12, 12, 13, 15, 16, 17,
		11, 11, 12, 12, 14, 15, 16, 17,
		12, 12, 13, 14, 15, 16, 17, 18,
		12, 12, 14, 16, 17, 18, 19, 19,
		13, 14, 15, 17, 19, 20, 20, 20,
		15, 15, 16, 18, 20, 21, 22, 22,
		16, 16, 17, 19, 21, 22, 23, 24,
		17, 17, 18, 19, 20, 22, 24, 24
	};

	// precompute Qp[] = quant * matrix
	for (int iii=0; iii<64; iii++) Qp[iii]=min(255,quant*qm[iii]);

}

SmoothD::~SmoothD() {

	// free akku
    free(akkuY);
    free(akkuU);
    free(akkuV);

}
	
PVideoFrame __stdcall SmoothD::GetFrame(int n, IScriptEnvironment* env) {

	PVideoFrame src = child->GetFrame(n, env);
	PVideoFrame dst = env->NewVideoFrame(vi);

	const int src_pitch = src->GetPitch();
	const int src_width = src->GetRowSize();
	const int src_height = src->GetHeight();

	const int dst_pitch = dst->GetPitch();
	const int dst_width = dst->GetRowSize();
	const int dst_height = dst->GetHeight();

	// check parameters for valid values
	if ((quant<1) || (quant>31))
		env->ThrowError("SmoothD: quant out of range");

	if ((num_shift<1) || (num_shift>8))
		env->ThrowError("SmoothD: num_shift out of range");
	
	if ((adaptive_shift<0) || (adaptive_shift>1))
		env->ThrowError("SmoothD: adaptive_shift out of range");
	
	if ((zero_weight<1) || (zero_weight>8))
		env->ThrowError("SmoothD: quant out of range");

	// handle adaptive_shift
	if (adaptive_shift==1) num_shift=2+quant/5;  // <- insert clever formula here

	//start processing frame
	if (vi.IsYUY2()) {
		// uh, oh!!
		// but ignore here
		// maybe convert in later version
	}

	if (vi.IsYV12()) {
		const unsigned char* srcp = src->GetReadPtr(PLANAR_Y);
		const unsigned char* srcpV = src->GetReadPtr(PLANAR_V);
		const unsigned char* srcpU = src->GetReadPtr(PLANAR_U);
		const int src_pitchUV = src->GetPitch(PLANAR_U);
		const int src_row_sizeUV = src->GetRowSize(PLANAR_U);
		const int src_heightUV = src->GetHeight(PLANAR_U);

		const int dst_pitchUV = dst->GetPitch(PLANAR_U);
		unsigned char* dstp = dst->GetWritePtr(PLANAR_Y);
		unsigned char* dstpV = dst->GetWritePtr(PLANAR_V);
		unsigned char* dstpU = dst->GetWritePtr(PLANAR_U);

		uint16_t *akkuYp, *akkuUp, *akkuVp;
		uint16_t *akkuY_srcp, *akkuU_srcp, *akkuV_srcp;

		int akkuY_pitch  = src_width;
		int akkuUV_pitch = src_row_sizeUV;

		int akkucount=0;	// number of values accumulated in akku

		// copy src to akku

		int h, x;
		akkuYp=akkuY;
		akkuUp=akkuU;
		akkuVp=akkuV;

		akkuY_srcp=akkuY_src;
		akkuU_srcp=akkuU_src;
		akkuV_srcp=akkuV_src;

		for (h=0; h < src_height; h++) {
			for (x=0; x < src_width; x++) {
				akkuYp[x]=((uint16_t)(srcp[x]))*zero_weight;
				akkuY_srcp[x]=(uint16_t)(srcp[x]);
			}
			srcp += src_pitch;
			akkuYp += akkuY_pitch;
			akkuY_srcp += akkuY_pitch;		// akkuY_src has same dimensions as akkuY
		}

		for (h=0; h < src_heightUV; h++) {
			for (x=0; x < src_width/2; x++) {
				akkuUp[x]=((uint16_t)(srcpU[x]))*zero_weight;
				akkuVp[x]=((uint16_t)(srcpV[x]))*zero_weight;
				akkuU_srcp[x]=(uint16_t)(srcpU[x]);
				akkuV_srcp[x]=(uint16_t)(srcpV[x]);
			}
			srcpU += src_pitchUV;
			srcpV += src_pitchUV;
			akkuUp += akkuUV_pitch;
			akkuVp += akkuUV_pitch;
			akkuU_srcp += akkuUV_pitch;		// akkuU_src has same dimensions as akkuU
			akkuV_srcp += akkuUV_pitch;		// akkuV_src has same dimensions as akkuV
		}

		akkucount=zero_weight;

		// shifts
		// 136
		// 248
		// 570
		// 408263715
		// 517362804

		int jblocks=2*((src_height+15)/16);
		int iblocks=2*((src_width +15)/16);
		
		int startj=8;	// just used for readability in DCT_Loop calls
		int starti=8;

		srcp = src->GetReadPtr(PLANAR_Y);
		srcpV = src->GetReadPtr(PLANAR_V);
		srcpU = src->GetReadPtr(PLANAR_U);

		switch(num_shift) {
		case 9:
			DCT_Loop(starti-5, startj-4, iblocks, jblocks);
			akkucount++;
		case 8:
			DCT_Loop(starti-1, startj-0, iblocks, jblocks);
			akkucount++;
		case 7:
			DCT_Loop(starti-7, startj-8, iblocks, jblocks);
			akkucount++;
      	case 6:
			DCT_Loop(starti-3, startj-2, iblocks, jblocks);
			akkucount++;
		case 5:
			DCT_Loop(starti-6, startj-6, iblocks, jblocks);
			akkucount++;
		case 4:
			DCT_Loop(starti-2, startj-3, iblocks, jblocks);
			akkucount++;
		case 3:
			DCT_Loop(starti-8, startj-7, iblocks, jblocks);
			akkucount++;
		case 2:
			DCT_Loop(starti-0, startj-1, iblocks, jblocks);
			akkucount++;
		case 1:
			DCT_Loop(starti-4, startj-5, iblocks, jblocks);
			akkucount++;
	    };

		// copy akku to destination

		akkuYp=akkuY;
		akkuUp=akkuU;
		akkuVp=akkuV;

		for (h=0; h < src_height; h++) {
			for (x=0; x < src_width; x++) {
				dstp[x]=(unsigned char)(akkuYp[x]/akkucount);
			}
			dstp += dst_pitch;
			akkuYp += akkuY_pitch;
		}

#ifndef PROCESS_UV
akkucount=zero_weight;
#endif

		for (h=0; h < src_heightUV; h++) {
			for (x=0; x < src_width/2; x++) {
				dstpU[x]=(unsigned char)(akkuUp[x]/akkucount);
				dstpV[x]=(unsigned char)(akkuVp[x]/akkucount);
			}
			dstpU += dst_pitchUV;
			dstpV += dst_pitchUV;
			akkuUp += akkuUV_pitch;
			akkuVp += akkuUV_pitch;
		}

	}

	if (!vi.IsYV12()) {
		env->ThrowError("SmoothD: YV12 only");
	}

	return dst;
}


void SmoothD::DCT_Loop(int starti, int startj, int iblocks, int jblocks) {

	int xx, yy, cy, cx;
	int h, x;
	short dct_block[64];
	int16_t coeff_block[64];

	uint16_t *akkuYp,     *akkuUp,     *akkuVp;
	uint16_t *akkuY_srcp, *akkuU_srcp, *akkuV_srcp;
	uint16_t *akkuY_tmpp, *akkuU_tmpp, *akkuV_tmpp;

	// shifted copy from akkuY_src to akkuY_tmp

	// blacken borders (currently does too much)
	akkuY_tmpp=akkuY_tmp;
	for (h=0; h < 8*jblocks; h++) {
		for (x=0; x < 8*iblocks; x++) {
			akkuY_tmpp[x]=0;  // should be 16 for CCIR
		}
		akkuY_tmpp += 8*iblocks;
	}

	akkuU_tmpp=akkuU_tmp;
	akkuV_tmpp=akkuV_tmp;

#ifdef PROCESS_UV
	for (h=0; h < 8*jblocks/2; h++) {
		for (x=0; x < 8*iblocks/2; x++) {
			akkuU_tmpp[x]=128;	// 128 == black
			akkuV_tmpp[x]=128;	// 128 == black
		}
		akkuU_tmpp += 8*iblocks/2;
		akkuV_tmpp += 8*iblocks/2;
	}
#endif

	akkuY_srcp=akkuY_src;
	akkuY_tmpp=akkuY_tmp;
	akkuY_tmpp += startj * 8*iblocks   + starti;

	for (h=0; h < vi.height; h++) {
		for (x=0; x < vi.width; x++) {
			akkuY_tmpp[x]=akkuY_srcp[x];
		}
		akkuY_srcp += vi.width;
		akkuY_tmpp += 8*iblocks;
	}

	akkuU_srcp=akkuU_src;
	akkuV_srcp=akkuV_src;
	akkuU_tmpp=akkuU_tmp;
	akkuV_tmpp=akkuV_tmp;
	akkuU_tmpp += startj * 8*iblocks/2 + starti/2;
	akkuV_tmpp += startj * 8*iblocks/2 + starti/2;
	
#ifdef PROCESS_UV
	for (h=0; h < vi.height/2; h++) {
		for (x=0; x < vi.width/2; x++) {
			akkuU_tmpp[x]=akkuU_srcp[x];
			akkuV_tmpp[x]=akkuV_srcp[x];
		}
		akkuU_srcp += vi.width/2;
		akkuV_srcp += vi.width/2;
		akkuU_tmpp += 8*iblocks/2;
		akkuV_tmpp += 8*iblocks/2;
	}
#endif

	// Process Y here
	for (yy=0; yy<jblocks; yy++) {
		for (xx=0; xx<iblocks; xx++) {

			// copy to dct_block
			akkuY_tmpp=akkuY_tmp; akkuY_tmpp += (yy*8)* iblocks*8 + xx*8;
			for (cy=0; cy<8; cy++) {
				for (cx=0; cx<8; cx++) dct_block[cy*8+cx]=akkuY_tmpp[cx];
				akkuY_tmpp += 8*iblocks;
			}

			// do DCT
			fdct_mmx_ffmpeg(dct_block);

			// do magic
			// this version has to do without magic

			// do quant
			quant_mpeg_intra_c(coeff_block, dct_block, quant, 1, quant_matrix);
			dequant_mpeg_intra_c(dct_block, coeff_block, quant, 1, quant_matrix);

			// do iDCT
			idct_mmx(dct_block);

			// copy back to akkuY_tmp
			akkuY_tmpp=akkuY_tmp; akkuY_tmpp += (yy*8)* iblocks*8 + xx*8;
			for (cy=0; cy<8; cy++) {
				for (cx=0; cx<8; cx++) akkuY_tmpp[cx]=dct_block[cy*8+cx];
				akkuY_tmpp += 8*iblocks;
			}
		}
	}

#ifdef PROCESS_UV
	// Process U and V here
	for (yy=0; yy<jblocks/2; yy++) {
		for (xx=0; xx<iblocks/2; xx++) {

			// copy to dct_block
			akkuU_tmpp=akkuU_tmp; akkuU_tmpp += (yy*8)* iblocks/2*8 + xx*8;
			for (cy=0; cy<8; cy++) {
				for (cx=0; cx<8; cx++) dct_block[cy*8+cx]=akkuU_tmpp[cx];
				akkuU_tmpp += 8*iblocks/2;
			}

			// do DCT
			fdct_mmx_ffmpeg(dct_block);

			// do magic
			// this version has to do without magic

			// do quant
			quant_mpeg_intra_c(coeff_block, dct_block, quant, 1, quant_matrix);
			dequant_mpeg_intra_c(dct_block, coeff_block, quant, 1, quant_matrix);

			// do iDCT
			idct_mmx(dct_block);

			// copy back to akkuU_tmp
			akkuU_tmpp=akkuU_tmp; akkuU_tmpp += (yy*8)* iblocks*8 + xx*8;
			for (cy=0; cy<8; cy++) {
				for (cx=0; cx<8; cx++) akkuU_tmpp[cx]=dct_block[cy*8+cx];
				akkuU_tmpp += 8*iblocks/2;
			}

			// copy to dct_block
			akkuV_tmpp=akkuV_tmp; akkuV_tmpp += (yy*8)* iblocks/2*8 + xx*8;
			for (cy=0; cy<8; cy++) {
				for (cx=0; cx<8; cx++) dct_block[cy*8+cx]=akkuV_tmpp[cx];
				akkuV_tmpp += 8*iblocks/2;
			}

			// do DCT
			fdct_mmx_ffmpeg(dct_block);

			// do magic
			// this version has to do without magic

			// do quant
			quant_mpeg_intra_c(coeff_block, dct_block, quant, 1, quant_matrix);
			dequant_mpeg_intra_c(dct_block, coeff_block, quant, 1, quant_matrix);

			// do iDCT
			idct_mmx(dct_block);

			// copy back to akkuV_tmp
			akkuV_tmpp=akkuV_tmp; akkuV_tmpp += (yy*8)* iblocks*8 + xx*8;
			for (cy=0; cy<8; cy++) {
				for (cx=0; cx<8; cx++) akkuV_tmpp[cx]=dct_block[cy*8+cx];
				akkuV_tmpp += 8*iblocks/2;
			}
		}
	}
#endif

	// shifted copy from akku_tmp to akku
	akkuYp=akkuY;
	akkuY_tmpp=akkuY_tmp;
	akkuY_tmpp += startj * 8*iblocks   + starti;

	for (h=0; h < vi.height; h++) {
		for (x=0; x < vi.width; x++) akkuYp[x] += akkuY_tmpp[x];
		akkuYp += vi.width;
		akkuY_tmpp += 8*iblocks;
	}

	akkuUp=akkuU;
	akkuVp=akkuV;
	akkuU_tmpp=akkuU_tmp;
	akkuV_tmpp=akkuV_tmp;
	akkuU_tmpp += startj * 8*iblocks/2 + starti/2;
	akkuV_tmpp += startj * 8*iblocks/2 + starti/2;

#ifdef PROCESS_UV
	for (h=0; h < vi.height/2; h++) {
		for (x=0; x < vi.width/2; x++) {
			akkuUp[x] += akkuU_tmpp[x];
			akkuVp[x] += akkuV_tmpp[x];
		}
		akkuUp += vi.width/2;
		akkuVp += vi.width/2;
		akkuU_tmpp += 8*iblocks/2;
		akkuV_tmpp += 8*iblocks/2;
	}
#endif

}


AVSValue __cdecl Create_SmoothD(AVSValue args, void* user_data, IScriptEnvironment* env) {
  PClip return_clip = new SmoothD(args[0].AsClip(), args[1].AsInt(3), args[2].AsInt(4), args[3].AsInt(0), args[4].AsInt(1), env);
  // Apply limiter after current filter.
  try {
	  AVSValue lim_args[1] = { return_clip };
	  return_clip = env->Invoke("Limiter", AVSValue(lim_args,1)).AsClip();
  } catch (AvisynthError) {}  // Just to be safe.

    return return_clip;
}


extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) {
	env->AddFunction("SmoothD", "c[quant]i[num_shift]i[adaptive_shift]i[zero_weight]i", Create_SmoothD, 0);
    return "`SmoothD' SmoothD plugin";
}
